#include <stdio.h>

// #include <__cross_studio_io.h>

#include "uforth.h"

#define printf debug_printf
#define getchar debug_getchar
#define gets debug_gets
#define feof debug_feof
#define exit debug_exit

struct dict *dict = (struct dict*)0xA000;


void stats() { 
  printf("\tDictionary: %d cells used out of %d  (%d percent used).\n", 
	 dict->here,  MAX_DICT_CELLS, (dict->here*100)/1024);
}


uforth_stat c_handle(void) {
  CELL r1 = dpop();
  char *str;
  switch(r1) {
  case 1: // dot 
    if (iram.didx > -1) {
      r1 = dpop();
      printf("%d ",r1);
      //      fflush(stdout);
    } else printf("EMPTY\n");
    break;
  case 2: // emit
    printf("%c", dpop()&0xff);
    break;
  case 3: // key
    dpush((CELL)getchar());
    break;
  case 4:
    stats();
    break;
  case 5:
    r1 = dpop();
    str = (char*)&(dict->d[dpop()]);
    printf("%.*s", r1, str);
    break;
  case 6:
    r1 = iram.didx;
    while (r1 >= 0) {
      printf("%d (0x%04x)\n",iram.ds[r1],iram.ds[r1]);
      --r1;
    }
    break;
  }

  return OK;
}

void create_c_handle_words() {
  uforth_interpret(": . 1 cf ;");
  uforth_interpret(": emit 2 cf ;");
  uforth_interpret(": key 3 cf ;");
  uforth_interpret(": mem 4 cf ;");
  uforth_interpret(": type 5 cf ;");
  uforth_interpret(": .s 6 cf ;");
}

static char line[128];
void interpret_from(void) {
  int16_t lineno = 0;
  while (1) {
    ++lineno;
    printf(" ok\n");
    gets(line);
    if (line[0] == '#' || line[0] == '\n') continue;

    switch(uforth_interpret(line)) {
    case E_NOT_A_WORD:
      printf("Huh? >>> %.*s <<< %s\n", iram.currwordlen, 
      	     iram.currword,
      	     iram.currword + iram.currwordlen);
      break;
    case E_ABORT:
      printf("Abort!\n");
      break;
    case E_STACK_UNDERFLOW:
      printf("Stack underflow!\n");
      break;
    case E_RSTACK_OVERFLOW:
      printf("Return Stack overflow!\n");
      break;
    case E_DSTACK_OVERFLOW:
      printf("Data Stack overflow!\n");
      break;
    case E_MISSING_LOOP:
      printf("Missing loop terminator!\n");
      break;
    case E_MISSING_THEN:
      printf("Missing then!\n");
      break;
    case OK:
      break;
    default:
      printf("Ugh\n");
      break;
    }
  }
}

int main( void )
{

  uforth_init();
  create_c_handle_words();

  stats();
  interpret_from();

  }
