#define RAM_DICT		/* use the RAM based dictionary */

#include <msp430x24x.h>
#include "uforth.h"
#include <string.h>

#ifdef RAM_DICT
struct dict ramdict = INITIAL_DICT;
struct dict *dict = &ramdict;
#else
# pragma constseg("DICT")
const struct dict flashdict = INITIAL_DICT;
# pragma constseg(default)
struct dict CONST *dict = &flashdict;

void dict_write(CELL idx, CELL cell) {
}

void dict_set_last_word(CELL cell) {
}

void dict_incr_here(CELL cell) {
}

void dict_append(CELL cell) {
}

void dict_append_string(char* src,CELL len) {
}
#endif /* RAM_DICT */

/* UART stuff */

#define IFG 		UC1IFG
#define UCRXIFG		UCA1RXIFG
#define UCTXIFG		UCA1TXIFG
#define UCRXBUF		UCA1RXBUF
#define UCTXBUF		UCA1TXBUF

uint8_t rxc(void) {
  while ((IFG & UCRXIFG) == 0);
  return UCRXBUF;
}

#define txc(c)   { while ((IFG & UCTXIFG) == 0);  UCTXBUF = c; }

void txs(char* s, int cnt) {
  while(cnt--) txc(*s++);
}

#define txs0(s) txs(s,strlen(s))

void rxgetline(char* str, uint8_t max) {
  char *ptr = str;
  while(max) {
    *ptr = rxc();
    if (*ptr == 3) { ptr = str; break; } /* ^C (abort) */
    if (*ptr == '\r' || *ptr == '\n') break;
    txc(*ptr);
    if (*ptr == 8) {		/* backspace */
      *ptr = '\0';
      if (ptr != str) { --ptr; --max; }
      continue; 
    }
    ++ptr; --max;
  }
  *ptr = '\0';
  txs0("\r\n");
}


void itoa(int16_t value, char* str, int16_t base);

void txn(uint16_t n) {
  static char num[10];
  itoa(n, num, 10);
  txs(num,strlen(num));
}

void stats(void) { 
  txs0("\tDict: ");
  txn(dict->here);
  txs0(" cells used out of ");
  txn(MAX_DICT_CELLS);
  txs0(" (");
  txn(((uint32_t)(dict->here))*100/MAX_DICT_CELLS);
  txs0("% used)\r\n");
}

uforth_stat c_handle(void) {
  CELL r1 = dpop();
  char *str;
  extern uforth_stat app_handle_cmds(CELL n);

  switch(r1) {
  case 1:			/* dot */
    if (iram.didx > -1) {
      r1 = dpop();
      txn(r1); txc(' ');
    } else txs0("EMPTY\r\n");
    break;
  case 2:			/* emit */
    txc(dpop()&0xff);
    break;
  case 3:			/* key */
    dpush((CELL)rxc());
    break;
  case 4:			/* stats */
    stats();
    break;
  case 5:			/* type */
    r1 = dpop();
    str = (char*)&(dict->d[dpop()]);
    txs(str,r1);
    break;
  case 6:			/* .s */
    r1 = iram.didx;
    while (r1 >= 0) {
      txn(iram.ds[r1]); txs0("\r\n");
      --r1;
    }
    break;
  default:
    return app_handle_cmds(r1);
  }
  return OK;
}

void create_c_handle_words(void) {
  extern void app_create_cmds(void);

  uforth_interpret(": . 1 cf ;");
  uforth_interpret(": emit 2 cf ;");
  uforth_interpret(": key 3 cf ;");
  uforth_interpret(": mem 4 cf ;");
  uforth_interpret(": type 5 cf ;");
  uforth_interpret(": .s 6 cf ;");

  app_create_cmds();
}

static char line[80];
void interpret_from(void) {
  int16_t lineno = 0;
  while (1) {
    ++lineno;
    txs0(" ok\r\n");
    rxgetline(line,79);
    if (line[0] == '#' || line[0] == '\r' || line[0] == '\0') continue;

    switch(uforth_interpret(line)) {
    case E_NOT_A_WORD:
      txs0("Huh? >>> ");
      txs(iram.currword,iram.currwordlen);
      txs0(" <<< ");
      txs0(iram.currword + iram.currwordlen);
      txs0("\r\n");
      break;
    case E_ABORT:
      txs0("Abort!\n");
      break;
    case E_STACK_UNDERFLOW:
      txs0("Stack underflow!\n");
      break;
    case E_DSTACK_OVERFLOW:
      txs0("Stack overflow!\n");
      break;
    case E_RSTACK_OVERFLOW:
      txs0("Return Stack overflow!\n");
      break;
    case E_MISSING_LOOP:
      txs0("Missing loop terminator!\n");
      break;
    case E_MISSING_THEN:
      txs0("Missing then!\n");
      break;
    case OK:
      break;
    default:
      txs0("Ugh\n");
      break;
    }
  }
}


int main( void )
{
  void app_init(void);

  WDTCTL = WDTPW+WDTHOLD;                   // Stop watchdog timer
  BCSCTL1 = CALBC1_1MHZ;                    // Set DCO
  DCOCTL = CALDCO_1MHZ;
  P3SEL = (1<<6)|(1<<7);	// P3.6,7 = USCI_A0 TXD/RXD
  UCA1CTL1 |= UCSSEL_2;                     // SMCLK
  UCA1BR0 = 104;                            // 1MHz 9600; (104)decimal = 0x068h
  UCA1BR1 = 0;                              // 1MHz 9600
  UCA1MCTL = UCBRS0;                        // Modulation UCBRSx = 1
  UCA1CTL1 &= ~UCSWRST;                     // **Initialize USCI state machine**

  uforth_init();

  app_init();

#ifdef BOOTSTRAP
  create_c_handle_words();
#endif
  stats();
  interpret_from();
}
